#include <stdlib.h>
#include "AdjList.h"

#define ACCURATE_LINKING  1

/* -----------------
 * Private Functions ======================================================
 * ----------------- 
 */

/* Given the head of a list, insert a new node with given id. 
 *
 * Returns boolean success of the operation.
 *
 * Two modes, one that checks whether the link already exists and
 * another that doesn't. Use the define statement at top to select
 * the mode (ACCURATE_LINKING).
 *
 * Not checking for pre-existing links means that the same node can
 * appear multiple times. However, this may be a small price to pay
 * for the potential speed increase of not having to traverse the link
 * list. 
 *
 * In ACCURATE_LINKING mode, the list is sorted by node id to reduce
 * the amount of traversal. Furthermore, the sort ordering is from
 * large to small since there is a correlation between the node ID and
 * when that node is used. (Completely correlated for Erdos-Renyi
 * graphs, less for others).
 */
int _adjlist_insert(AdjNode* head, int id) {
  AdjNode *new, *pre, *cur;

  /* check if list is empty */
  if (head->id == UNSET_ADJNODE) {
    head->id = id;        /* set the head to be the input value */
    return(1);
  }

  pre = cur = head;
  if (ACCURATE_LINKING) {
    /* find where to insert */
    while (cur != 0  &&  cur->id > id) {
      pre = cur;
      cur = cur->next;
    }

    /* check if this link already exists */
    if (cur != 0  &&  cur->id == id)
      return(0);
  }
  
  /* create the new link */
  new = malloc(sizeof(AdjNode));
  /* take care not to link to the previous node */
  if (cur == pre) { /* always true for non accurate linking */
    /* copy the head data into our new node */
    new->id   = pre->id;
    new->next = pre->next;  // == cur
    /* set the new head data */
    pre->id   = id;
    pre->next = new;
  }
  else {            /* only used in accurate linking mode */
    new->id   = id;
    new->next = cur;
    pre->next = new;
  }

  return(1);
}
       

/* Creates an edge betwen to nodes. */
void _adjlist_link(AdjList* adjlist, int id1, int id2) {
  /* no self linking */
  if (id1 == id2)
    return;
  
  /* undirected graph, so make each the neighbor of the other */
  if (_adjlist_insert(adjlist->list[id1], id2))
    adjlist->degree[id1]++;
  if (_adjlist_insert(adjlist->list[id2], id1))
    adjlist->degree[id2]++;
}


/* Clears the adjacency list. */
void _adjlist_init(AdjList* adjlist) {
  AdjNode *node, *nuke;
  int y = 0;
  while (y < adjlist->size) {
    /* get first node for this neighbor list */
    node = adjlist->list[y++];
    /* set it to the pristine state and move on */
    node->id = UNSET_ADJNODE;
    nuke = node;
    node = node->next;
    nuke->next = 0;

    /* free the rest of the neighbors */
    while (node) {
      nuke = node;
      node = node->next;
      free(nuke);
    }
  }
}


/* returns boolean based on whether the two nodes are linked */
int _adjlist_isLinked(AdjList* adjlist, int id1, int id2) {
  int       lrg, sml;
  AdjNode*  cur;

  /* no self-linking, return false */
  if (id1 == id2)
    return (0);

  /* since we sort biggest to largest, we should try to find whether
   * the larger id is in the smaller id's list. */
  ( id1 > id2 ) ? (lrg = id1, sml = id2) : (lrg = id2, sml = id1);
  
  /* traverse through all the larger IDs */
  for (cur = adjlist->list[sml] ; 
       cur != 0  &&  cur->id > lrg ;
       cur = cur->next) ;
  
  if (cur && cur->id == lrg)
    return (1);
  return(0);
}


  

/* ----------------
 * Public Functions =======================================================
 * ---------------- 
 */

AdjList newAdjList(int size) { 
  int i;
  AdjList adjlist;
  
  adjlist.size     = size;
  adjlist.list     = malloc(adjlist.size * sizeof(AdjNode*));
  for (i = 0; i < adjlist.size; i++) {
    adjlist.list[i] = malloc(sizeof(AdjNode));
    adjlist.list[i]->id = UNSET_ADJNODE;
  }
  adjlist.degree   = calloc(size, sizeof(int));
  adjlist.link     = _adjlist_link;
  adjlist.init     = _adjlist_init;
  adjlist.isLinked = _adjlist_isLinked;
  return adjlist;
}

