#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "AdjList.h"
#include "Graph.h"
#include "measure.h"
#include "randgraphs.h"

#ifdef OSX
#include <string.h>
#endif

#define MEASURE_C_DEBUG 0

/* -----------------
 * Private Functions ======================================================
 * ----------------- 
 */


/* A modified version of measGiantComponent(), this
 * returns the node id of the leader. */
int _getGiantSourceNode(Graph G) {
  int i, node = -1, giant = 0;
  assert(G.hasUnionFind);
  for (i = 0; i < G.unionfind.size; i++)
    if (G.unionfind.array[i] < giant) {
      giant = G.unionfind.array[i];
      node  = i;
    }
  return(node);
}


/* Returns a list of all nodes that are part of the giant component. */
int* _getGiantNodeList(Graph G) {
  int  i, idx = 0, source = _getGiantSourceNode(G);
  int *list = malloc((G.unionfind.size + 1) * sizeof(int));

  /* insert the source node */
  list[idx++] = source;
  
  /* find and insert the giant component nodes */
  for (i = 0; i < G.unionfind.size; i++)
    if (G.unionfind.array[i] == source)
      list[idx++] = i;

  /* mark the end of the list */
  list[idx] = -1;

  return (list);
}


/* Returns the average pathlength of a BFS tree for passed graph and
 * source node.
 *
 * A note on my implementation of BFS: I simulate a fifo queue by
 * doubly indexing an array. One indice points to the beginning of the
 * list and one points to the end. Equality between indices indicates
 * an empty queue. To add data, put it at the tail and increment the
 * tail. To remove data, increment the head index. 
 */
double _PathLenBFS(Graph G, int source) {
  int i, qhead, qtail, dist, nodecnt, parent, total;
  int *Q, *d;
  AdjNode *nbrNode;
  
  assert(G.hasList);
  
  Q = malloc(G.adjlist.size * sizeof(int));
  d = calloc(G.adjlist.size,  sizeof(int));
  qhead = qtail = dist = 0;
  
  /* process source node */
  Q[qtail++]  = source;
  d[source] = dist;

  /* process the queue until it empties out */
  nodecnt = -1;
  while (qhead != qtail) {
    nodecnt++;
    parent = Q[qhead];
    nbrNode = G.adjlist.list[Q[qhead++]];
    while (nbrNode) {
      /* process node if we haven't seen it before */
      if (! d[nbrNode->id] && nbrNode->id != source) {
        d[nbrNode->id] = d[parent] + 1;
        Q[qtail++] = nbrNode->id;
      }
      nbrNode = nbrNode->next;
    }
  }

  total = 0;
  for (i = 0; i < G.size; i++)
    total += d[i];
  
  free(Q);
  free(d);

  /* single node tree has length 0 */
  if (nodecnt <= 0)
    return 0;
  
  return (total / (double)nodecnt);
}
    

/* Returns the clustering coefficient using the
 * input graph's adjacency matrix.
 */
double _measAvgClustCoeff_matrix(Graph G) { 
  double  coeff, runningtotal = 0;
  int     n, nodecnt = 0;
  int     *nbrs, nbrcnt, nbr;
  int     edgecnt;
  int     i, j, x, y, tmp;
  
  /* measure clustering coefficient for each vertex */
  for (j = 0; j < G.adjmat.size; j++) {

    if (MEASURE_C_DEBUG)
      printf("\nNode %d\n", j);
    
    /* count neighbors of vertex j */
    nbrcnt = 0;
    for (i = 0; i < G.adjmat.size; i++) {
      (i < j) ? (y = j, x = i) : (y = i, x = j);
      if (G.adjmat.matrix[y][x])
        nbrcnt++;
    }
    if (MEASURE_C_DEBUG)
      printf("Node %d has %d neighbors\n", j, nbrcnt);
    if (nbrcnt < 2)
      continue;

    /* create and populate an array of the neighbors */
    nbrs = malloc(nbrcnt * sizeof(int));
    for (nbr = 0, i = 0; i < G.adjmat.size; i++) {
      (i < j) ? (y = j, x = i) : (y = i, x = j);
      if (G.adjmat.matrix[y][x])
        nbrs[nbr++] = i;
    }
    if (MEASURE_C_DEBUG) {
      printf("Node %d has neighbors: ",j);
      for (tmp = 0; tmp < nbrcnt; tmp++)
        printf("%d ", nbrs[tmp]);
      printf("\n");
    }

    /* determine connectedness of the nbrs */
    edgecnt = 0;
    for (nbr = 0; nbr < nbrcnt; nbr++) {
      for (n = nbr + 1; n < nbrcnt; n++) {
        (nbrs[nbr] < nbrs[n]) ? 
          (y = nbrs[n], x = nbrs[nbr]) : 
          (y = nbrs[nbr], x = nbrs[n]) ;
        if (G.adjmat.matrix[y][x])
          edgecnt++;
      }
    }
    free(nbrs);
    if (MEASURE_C_DEBUG)
      printf("Node %d neighbors have %d edges to each other\n", j, edgecnt);

    /* calculate the clustering coefficient for this vertex */
    /* coeff = edgecnt / C(N,2)  */
    /* C(N,2) = N * (N - 1) / 2  */
    coeff = edgecnt / (nbrcnt * (nbrcnt - 1) / 2.0);
    runningtotal += coeff;
    nodecnt++;

    if (MEASURE_C_DEBUG)
      printf("clust coeff for node %d is %f\n", j, coeff);
  }
 
  return (runningtotal / (double)nodecnt);
}


/* Returns the clustering coefficient of the input graph's
 * giant component using the graph's adjacency matrix.
 */
double _measAvgCompClustCoeff_matrix(Graph G) { 
  double  coeff, runningtotal = 0;
  int     n, nodecnt = 0;
  int     *nbrs, nbrcnt, nbr;
  int     edgecnt;
  int     i, j, k, x, y, tmp;
  int     *complist;

  /* measure clustering coefficient for each 
   * vertex in the giant component */
  complist = _getGiantNodeList(G);
  for (k = 0, j = complist[k]; j != -1; j = complist[++k]) {

    if (MEASURE_C_DEBUG)
      printf("\nNode %d\n", j);
    
    /* count neighbors of vertex j */
    nbrcnt = 0;
    for (i = 0; i < G.adjmat.size; i++) {
      (i < j) ? (y = j, x = i) : (y = i, x = j);
      if (G.adjmat.matrix[y][x])
        nbrcnt++;
    }
    if (MEASURE_C_DEBUG)
      printf("Node %d has %d neighbors\n", j, nbrcnt);
    if (nbrcnt < 2)
      continue;

    /* create and populate an array of the neighbors */
    nbrs = malloc(nbrcnt * sizeof(int));
    for (nbr = 0, i = 0; i < G.adjmat.size; i++) {
      (i < j) ? (y = j, x = i) : (y = i, x = j);
      if (G.adjmat.matrix[y][x])
        nbrs[nbr++] = i;
    }
    if (MEASURE_C_DEBUG) {
      printf("Node %d has neighbors: ",j);
      for (tmp = 0; tmp < nbrcnt; tmp++)
        printf("%d ", nbrs[tmp]);
      printf("\n");
    }

    /* determine connectedness of the nbrs */
    edgecnt = 0;
    for (nbr = 0; nbr < nbrcnt; nbr++) {
      for (n = nbr + 1; n < nbrcnt; n++) {
        (nbrs[nbr] < nbrs[n]) ? 
          (y = nbrs[n], x = nbrs[nbr]) : 
          (y = nbrs[nbr], x = nbrs[n]) ;
        if (G.adjmat.matrix[y][x])
          edgecnt++;
      }
    }
    free(nbrs);
    if (MEASURE_C_DEBUG)
      printf("Node %d neighbors have %d edges to each other\n", j, edgecnt);

    /* calculate the clustering coefficient for this vertex */
    /* coeff = edgecnt / C(N,2)  */
    /* C(N,2) = N * (N - 1) / 2  */
    coeff = edgecnt / (nbrcnt * (nbrcnt - 1) / 2.0);
    runningtotal += coeff;
    nodecnt++;

    if (MEASURE_C_DEBUG)
      printf("clust coeff for node %d is %f\n", j, coeff);
  }
 
  return (runningtotal / (double)nodecnt);
}


double _measAvgClustCoeff_list(Graph G) {
  int       i, nbrcnt, edgecnt, nodecnt;
  double    total = 0.0;
  AdjNode   *cur1, *cur2;

  printf("\nWARNING: Calculating clustering coefficient on adjacency lists is\n");
  printf("         very slow. Use the matrix representation if memory allows.\n");
  
  nodecnt = 0;
  for (i = 0; i < G.adjlist.size; i++) {

    /* undefined for vertices with less than two neighbors */
    if (G.adjlist.degree[i] < 2)
      continue;
    
    edgecnt = nbrcnt = 0;
    for (cur1 = G.adjlist.list[i]; 
         cur1 != 0  &&  cur1->id != UNSET_ADJNODE; 
         cur1 = cur1->next) {
      nbrcnt++;
      for (cur2 = cur1->next; cur2 != 0; cur2 = cur2->next)
        if (G.adjlist.isLinked(&G.adjlist, cur1->id, cur2->id))
          edgecnt++;
    }
 
    /* calculate the clustering coefficient for this vertex */
    /*   coeff  = edgecnt / C(N,2)                          */
    /*   C(N,2) = N * (N - 1) / 2                           */
    total += edgecnt / (nbrcnt * (nbrcnt - 1) / 2.0);
    nodecnt++;
  }

  return (total / (double)nodecnt);
}

/* ----------------
 * Public Functions =======================================================
 * ---------------- 
 */

/* Modifies a passed integer array of degree distribution
 * for the input graph.
 *
 * Array is indexed by degree and stores the number
 * of nodes of that degree.
 */
void measDegDist(Graph G, int *dist) {
  int i, x, y;
  int *degree;

  /* adjacency list "calculation" */
  if (G.hasList) {
    for (i = 0; i < G.size; i++)
      dist[G.adjlist.degree[i]]++;
  }
  
  /* adjacency matrix calculation */
  else if (G.hasMatrix) {
    /* initialize degree vector */
    degree = calloc(G.size, sizeof(int));
    
    /* calculate the degree of each node */
    for (y = 1; y < G.size; y++)
      for (x = 0; x < y; x++) {
        degree[x] += G.adjmat.matrix[y][x];
        degree[y] += G.adjmat.matrix[y][x];
      }
    
    /* calculate the degree distribution */
    for (i = 0; i < G.size; i++)
      dist[degree[i]]++;
  }

  else
    printf("\n!!ERROR!! No valid data structure for computing degree distribution!\n");

}


/* Modifies a passed integer array of degree distribution
 * for the input graph's giant component.
 *
 * Array is indexed by degree and stores the number
 * of nodes of that degree.
 */
void measCompDegDist(Graph G, int *dist) {
  int *nodelist, node, i;
  nodelist = _getGiantNodeList(G);
  for (i = 0; (node = nodelist[i]) != -1; i++)
    dist[G.adjlist.degree[node]]++;
}

double measAvgClustCoeff(Graph G) {
  if (G.hasMatrix)
    return _measAvgCompClustCoeff_matrix(G);
  else if (G.hasList)
    return _measAvgClustCoeff_list(G);
  printf("\n!!ERROR!! No valid data structure for computing clustering coefficient!\n");
  return -42;
}


/* Returns the size of the giant component. */
int measGiantComponent(Graph G) {
  int i, giant = 0;
  for (i = 0; i < G.unionfind.size; i++)
    if (G.unionfind.array[i] < giant)
      giant = G.unionfind.array[i];
  return(-giant);
}


/* Returns the size of the giant component averaged
 * over multiple runs. */
double measAvgGiantComponent(Graph G, Opts* opts) {
  int i, s, total, giant=0;

  opts->G = &G;   /* same graph will be iterated over */

  /* our first giant */
  for (i = 0; i < G.unionfind.size; i++) {
    if (G.unionfind.array[i] < giant)
      giant = G.unionfind.array[i];
    G.unionfind.array[i] = -1;              /* reset for next run */
  }
  total = -giant;
  
  for (s = 1; s < opts->samples; s++) {
    genGraph(opts);                       /* iterate graph */
    for (i = 0; i < G.unionfind.size; i++) {
      if (G.unionfind.array[i] < giant)
        giant = G.unionfind.array[i];
      G.unionfind.array[i] = -1;            /* reset for next run */
    }
    total -= giant;
  }
  
  return (total / (double)opts->samples / (double)G.size);
}


/* Calculates the average pathlength between nodes in the giant component
 * using my own BFS to calculate shortest paths. */
double measAvgPathLen(Graph G) { 
  int     i, *nodelist = _getGiantNodeList(G);
  double  total = 0;
  
  for (i = 0; nodelist[i] != -1 ; i++) {
    total += _PathLenBFS(G, nodelist[i]);
  }

  return(total / (double)i);
}

