#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "randgraphs.h"

#ifdef OSX
#include <string.h>
#endif


/* -----------------
 * Private Functions ======================================================
 * ----------------- 
 */

Graph _genEmpty(Opts *opts) {
  return newGraph(opts);
}


/* Generates an Erdos-Renyi random graph using the binomial method.
 * 
 * Model: Each node connects to every other node with probability p.
 */
Graph _genErdos(Opts *opts) {
  int   pt1, pt2;
  long  p;
  Graph G;

  /* check if we are modifying an existing graph */
  if (opts->G)
    G = *opts->G;
  else
    G = newGraph(opts);

  assert(opts->prob <= 1.0);
  p = opts->prob * RAND_MAX;

  for (pt1 = 0; pt1 < opts->size - 1; pt1++)
    for (pt2 = pt1 + 1; pt2 < opts->size; pt2++)
      if (random() < p)
          G.link(&G, pt1, pt2);

  return G;
}

/* Generates a slightly modified Watts-Strogatz model of a small world
 * network. The model is modified to have new edges enter the graph
 * without replacing the old.
 */
Graph _genWattsStrogatz(Opts *opts) {
  int   node, k;
  long  p;
  Graph G;

  /* check if we are modifying an existing graph */
  if (opts->G)
    G = *opts->G;
  else
    G = newGraph(opts);

  assert(opts->links > 0);
  assert(opts->prob <= 1.0);
  p = opts->prob * RAND_MAX;

  /* generate locally connected graph (opts->links nearest neighbors) */
  for (node = 0; node < opts->size ; node++)
    for (k = 1 ; k <= opts->links ; k++) 
      G.link(&G, node, (node + k) % opts->size);

  /* randomly wire the graph */
  opts->G = &G;
  _genErdos(opts);

  return G;
}

/* Generates the random graph described in the paper "Are Random Graphs
 * Really Random?" by Callaway, Hopcroft, Kleinberg, Newman, and Strogatz,
 * 2001 <URL:http://www.santafe.edu/sfi/publications/wpabstract/200105025>.
 *
 * Model: At each time step, a new vertex is added. Then, with
 *        probability p, two vertices are chosen uniformly at random
 *        and joined by an undirected edge.
 *
 *        Initializing the graph with a single isolated vertex does
 *        not change its asymptotic behavior.
 */
Graph _genCHKNS(Opts *opts) {
  int   i, rn1, rn2;
  long  p;
  Graph G;

  /* check if we are modifying an existing graph */
  if (opts->G)
    G = *opts->G;
  else
    G = newGraph(opts);

  assert(opts->prob <= 1.0);
  p = opts->prob * RAND_MAX;

  for (i = 1; i < opts->size ; i++) {
    if (random() < p) {           /* with probability p, */
      rn1 = random() % i;     /* choose random vertices from those */
      rn2 = random() % i;     /* "already added" and               */
                              /* join them */
      G.link(&G, rn1, rn2);
    }
  }

  return G;
}


/* David Lancaster growth models
 *
 * A generalized algorithm for generating the two random graphs described in
 * the paper "Cluster growth in two growing network models" by David
 * Lancaster <URL:http://stacks.iop.org/JPhysA/35/1179>:
 *
 * Model: In each time step a new vertex is introduced. With probability p,
 *        the new vertex is connected to opts->links other vertices, chosen
 *        at random from amongst the existing vertices. The vertex remains
 *        disconnected with probability 1 - p. In numerical simulations we
 *        always start with a single intial vertex under the assumption
 *        this initial condition won't affect results at large times. 
 *
 *        It differs from the CHKNS model above since only the new vertex
 *        has an opportunity to make a link.
 *
 * Notes: For p < 1 and opts->links < 2, this process creates a forrest of
 *        trees and does not percolate.
 */
Graph _genLancaster(Opts *opts) {
  int   n, l, rn;
  long  p;
  Graph G;

  /* check if we are modifying an existing graph */
  if (opts->G)
    G = *opts->G;
  else
    G = newGraph(opts);

  assert(opts->links > 0);
  assert(opts->prob <= 1.0);
  p = opts->prob * RAND_MAX;

  for (n = 1; n < opts->size ; n++) {
    /* with probability p, link to a random existing vertex */
    if (random() < p) {
      for (l = 0; l < opts->links; l++) {
        rn = random() % n;   
        G.link(&G, n, rn);
      }
    }
  }

  return G;
}



/* ----------------
 * Public Functions =======================================================
 * ---------------- 
 */


Graph genGraph(Opts *opts) {

  if (strcmp(opts->type, "chkns") == 0) {
    return _genCHKNS(opts);
  }
  else if (strcmp(opts->type, "empty") == 0) {
    return _genEmpty(opts);
  }
  else if (strcmp(opts->type, "erdos") == 0) {
    return _genErdos(opts);
  }    
  else if (strcmp(opts->type, "watts") == 0) {
    return _genWattsStrogatz(opts);
  }
  else if (strcmp(opts->type, "lancaster") == 0) {
    return _genLancaster(opts);
  }
  else {
    printf("!! No graph type specified! Generating empty graph.\n");
    return _genEmpty(opts);
  }

}
